/*==================================================================*/
/*
 * Parallel Recursive Spectral Bisection -
 * Based on the sequential version provided by Horst Simon
 * -ravi 
 */
/*==================================================================*/

#include "parti.h"
#include <stdio.h>

#define UNUSED -1000
#define FIEDLER

/*
#define DEBUG
#include <util.h>
#define malloc(size) NEWN(char, size)
#define free(ptr)    DISPOSE(ptr)
*/

/*==================================================================*/
/* the spectral partitioner begins here, it takes input in CSR format 
* xadj_ind: the node (row) index 
* xadj: the row ptr in CSR
* adj:  the adjacency list 
* neqns: number of nodes assigned to me 
* nadj: number of adjacent nodes for my nodes 
* pfac: partition factor
* -ravi 08/05/91
*/
/*==================================================================*/

void genfid(xadj_ind, neqns, xadj, adj,  nadj, pfac, color)
int *xadj_ind, *xadj, *adj;
int neqns, nadj, pfac;
int *color;

{

  TTABLE *ttable, *ntable;
  SCHED  *ssched, *nsched;
  int i, j, k, l;
  int npart;
  int cur_color;
  int *xadloc_ind, *xadloc, *adloc;
  int *perm;
  int nodloc, nadloc;
  double *fied;
  double dwork;
  int *icomp;
  int *work;
  int *sorder;
  int half;
  int error;
  int *icolor;
  
  int *loc_ref, *float_color;
  int teqns;
  int nfied;
  int half1, half2;
  char fname[10];
  FILE *fp;

  int *localized_col, n_off_proc;
  int  n_off_proc1;
  double st_time, time;
  /* initialize */

#ifdef DEBUG
  printf("entered genfid ndata %d nadj %d pfac %d\n", neqns, nadj, pfac); gsync();
#endif

  xadloc_ind = NULL;
  xadloc = NULL;
  adloc = NULL;

  npart  = (1 << pfac);
  icomp  = (int *) malloc(sizeof(int)*npart);
  work   = (int *) malloc(sizeof(int)*npart);
  fied   = (double *) malloc(sizeof(double)*neqns);
  sorder = (int *) malloc(sizeof(int)*neqns);

  
  icomp[0] = neqns;
  PARTI_gisum(&icomp[0], 1, &work[0]);
  for (i=1; i<npart; i++) {
    icomp[i] = 0;
  }
 
  ttable        = (TTABLE *) build_translation_table(1, xadj_ind, neqns);
  localized_col = (int *) malloc(sizeof(int)*nadj);
  localize(ttable, &nsched, adj, localized_col, nadj, &n_off_proc1, neqns, 1);
  icolor        = (int *) malloc(sizeof(int)*(neqns+n_off_proc1));

  for(i=0; i<neqns; i++) {
    icolor[i] = 0;
  }

  /* loop over npart times */
  for(nfied=0,i=0; i<pfac; i++) {
     for(j=0; j<(1<<i); j++) {
       cur_color = j;

       PARTI_gsync();


/*
       if(xadloc_ind != NULL) free(xadloc_ind);
       if(xadloc != NULL) free(xadloc);
       if(adloc != NULL) free(adloc);
*/
       renumber(ttable, nsched,xadj_ind, xadj, adj, localized_col, icolor, 
         cur_color, neqns, nadj, n_off_proc1,
         &perm, &xadloc_ind, &xadloc, &adloc, &nadloc, &nodloc, &teqns);


#ifdef FIEDLER


       if (PARTI_mynode == 0) {printf("(%d) \t ",  nfied++);fflush(stdout);}

#endif
       ntable = (TTABLE *)build_translation_table(1, xadloc_ind, nodloc);
       genord(ntable, xadloc_ind, xadloc, adloc, nodloc, nadloc,  fied, teqns);
       crib_sort(fied, nodloc, sorder);
       sprintf(fname, "FIED.%d", PARTI_mynode);
       fp = fopen(fname, "w");
       for(l=0; l<nodloc; l++) {
         fprintf(fp, "%d -- %f \n", l, fied[l]);
       }
       fclose(fp);
       half = icomp[j] / 2;

       /* update the original color */

       
       half1 = half2 = 0;

       loc_ref = (int *) malloc(sizeof(int)*nodloc);
       localize(ntable, &ssched, sorder, loc_ref, nodloc,&n_off_proc, nodloc,1);
       float_color = (int *) malloc(sizeof(int)*(nodloc+n_off_proc));
       for(l=0; l<nodloc+n_off_proc; l++) {
          float_color[l] = UNUSED;
       }

       for(k=0; k<nodloc; k++) {
         if(xadloc_ind[k] < half) {
           float_color[loc_ref[k]] = -(2*j);
           half1++;
         }
         else {
           float_color[loc_ref[k]] = -(2*j)-1;
           half2++;
         }
       }
       new_scatter(ssched, &float_color[nodloc], float_color, sizeof(int), NULL);
       for(l=0; l<nodloc; l++) {
         if (float_color[l] != UNUSED) {
           icolor[perm[l]] = float_color[l];
         }
       }
       free(float_color);free(loc_ref); 
       free_sched(ssched); 

/*
       for(l=0; l<nodloc; l++) {
         if(sorder[l] < half) {
            icolor[perm[l]] = -(2*j);
         }
         else
           icolor[perm[l]] = -(2*j+1);
       }

*/


       /* memory allocated in renumber() */
       free(xadloc); free(adloc); 
       free(xadloc_ind); free(perm);


       free_table(ntable); 


     }
     for(k=0; k<neqns; k++) {
       icolor[k] = -icolor[k];
     } 
     for (l=0; l<npart; l++) {
       icomp[l] = 0;
     }

     for(l=0; l<neqns; l++) {
       icomp[icolor[l]]++;
     }

     PARTI_gisum(icomp, npart, work);

/*
    printf("-->AFTER GSUM ICOMP %d %d %d \n", icomp[0], icomp[1], icomp[2]);
*/
  }
  free(localized_col);
  free(sorder);
  free(icomp);
  free(work);
  free(fied);
  free_sched(nsched);
 
  for(i=0; i<neqns; i++) color[i] = icolor[i];
  free(icolor);

}
/*==================================================================*/

void eigen_partitioner_(xadj_ind, neqns, xadj, adj, nadj, pfac, color)
int *xadj_ind, *xadj, *adj;
int *neqns, *nadj, *pfac;
int *color;
{
    int neqns_val, nadj_val, pfac_val;

    neqns_val = *neqns;
    nadj_val  = *nadj;
    pfac_val  = *pfac;

    genfid(xadj_ind, neqns_val, xadj, adj, nadj_val, pfac_val, color);
}
/*==================================================================*/

void eigen_partitioner(xadj_ind, neqns, xadj, adj,  nadj, pfac, color)
int *xadj_ind, *xadj, *adj;
int neqns, nadj, pfac;
int *color;
{

    genfid(xadj_ind, neqns, xadj, adj, nadj, pfac, color);
}

/*==================================================================*/
